// Field.cpp: implementation of the Field class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Field.h"

//////////////////////////////////////////////////////////////////////
// Squares
//////////////////////////////////////////////////////////////////////
const int NIL = (Square::NIL);
const int lru = (Square::L | Square::R | Square::U);
const int lr  = (Square::L | Square::R);
const int lrd = (Square::L | Square::R | Square::D);
const int lud = (Square::L | Square::U | Square::D);
const int rud = (Square::R | Square::U | Square::D);
const int ud  = (Square::U | Square::D);
const int ld  = (Square::L | Square::D);
const int rd  = (Square::R | Square::D);
const int ru  = (Square::R | Square::U);
const int lu  = (Square::L | Square::U);
const int u   = (Square::U);
const int d   = (Square::D);
const int r   = (Square::R);
const int l   = (Square::L);

//////////////////////////////////////////////////////////////////////
// 1. Piece O
//////////////////////////////////////////////////////////////////////
int PieceOSquares[] =
{
	NIL, NIL, NIL, NIL,
	NIL,  lu,  ru, NIL,
	NIL,  ld,  rd, NIL,
	NIL, NIL, NIL, NIL,
};

Piece PieceO(Mod(0, 160, 255), 1, 4, 4, PieceOSquares);

//////////////////////////////////////////////////////////////////////
// 2. Piece I
//////////////////////////////////////////////////////////////////////
int PieceISquares[] =
{
	NIL, NIL, NIL, NIL,
	lud,  ud,  ud, rud,
	NIL, NIL, NIL, NIL,
	NIL, NIL, NIL, NIL,

	NIL, NIL, lru, NIL,
	NIL, NIL,  lr, NIL,
	NIL, NIL,  lr, NIL,
	NIL, NIL, lrd, NIL,
};

Piece PieceI(Mod(220, 64 + 32, 20 + 64), 2, 4, 4, PieceISquares);

//////////////////////////////////////////////////////////////////////
// 3. Piece S
//////////////////////////////////////////////////////////////////////
int PieceSSquares[] =
{
	NIL, NIL, NIL, NIL,
	NIL, NIL,  lu, rud,
	NIL, lud,  rd, NIL,
	NIL, NIL, NIL, NIL,

	NIL, NIL, lru, NIL,
	NIL, NIL, ld,  ru,
	NIL, NIL, NIL, lrd,
	NIL, NIL, NIL, NIL,
};

Piece PieceS(Mod(255, 160, 0), 2, 4, 4, PieceSSquares);

//////////////////////////////////////////////////////////////////////
// 4. Piece Z
//////////////////////////////////////////////////////////////////////
int PieceZSquares[] =
{
	NIL, NIL, NIL, NIL,
	NIL, lud,  ru, NIL,
	NIL, NIL,  ld, rud,
	NIL, NIL, NIL, NIL,

	NIL, NIL, NIL, lru,
	NIL, NIL,  lu,  rd,
	NIL, NIL, lrd, NIL,
	NIL, NIL, NIL, NIL,
};

Piece PieceZ(Mod(255, 255, 128), 2, 4, 4, PieceZSquares);

//////////////////////////////////////////////////////////////////////
// 5. Piece L
//////////////////////////////////////////////////////////////////////
int PieceLSquares[] =
{
	NIL, NIL, NIL, NIL,
	NIL,  lu,  ud, rud,
	NIL, lrd, NIL, NIL,
	NIL, NIL, NIL, NIL,

	NIL, NIL, lru, NIL,
	NIL, NIL,  lr, NIL,
	NIL, NIL,  ld, rud,
	NIL, NIL, NIL, NIL,

	NIL, NIL, NIL, lru,
	NIL, lud,  ud,  rd,
	NIL, NIL, NIL, NIL,
	NIL, NIL, NIL, NIL,

	NIL, lud,  ru, NIL,
	NIL, NIL,  lr, NIL,
	NIL, NIL, lrd, NIL,
	NIL, NIL, NIL, NIL,
};

Piece PieceL(Mod(0, 192, 192), 4, 4, 4, PieceLSquares);

//////////////////////////////////////////////////////////////////////
// 6. Piece J
//////////////////////////////////////////////////////////////////////
int PieceJSquares[] =
{
	NIL, NIL, NIL, NIL,
	NIL, lud,  ud,  ru,
	NIL, NIL, NIL, lrd,
	NIL, NIL, NIL, NIL,

	NIL, NIL,  lu, rud,
	NIL, NIL,  lr, NIL,
	NIL, NIL, lrd, NIL,
	NIL, NIL, NIL, NIL,

	NIL, lru, NIL, NIL,
	NIL,  ld,  ud, rud,
	NIL, NIL, NIL, NIL,
	NIL, NIL, NIL, NIL,

	NIL, NIL, lru, NIL,
	NIL, NIL,  lr, NIL,
	NIL, lud,  rd, NIL,
	NIL, NIL, NIL, NIL,
};
Piece PieceJ(Mod(32, 255, 160), 4, 4, 4, PieceJSquares);

//////////////////////////////////////////////////////////////////////
// 7. Piece T
//////////////////////////////////////////////////////////////////////
int PieceTSquares[] =
{
	NIL, NIL, NIL, NIL,
	NIL, lud,   u, rud,
	NIL, NIL, lrd, NIL,
	NIL, NIL, NIL, NIL,

	NIL, NIL, lru, NIL,
	NIL, NIL,   l, rud,
	NIL, NIL, lrd, NIL,
	NIL, NIL, NIL, NIL,

	NIL, NIL, lru, NIL,
	NIL, lud,   d, rud,
	NIL, NIL, NIL, NIL,
	NIL, NIL, NIL, NIL,

	NIL, NIL, lru, NIL,
	NIL, lud,   r, NIL,
	NIL, NIL, lrd, NIL,
	NIL, NIL, NIL, NIL,
};

Piece PieceT(Mod(192, 64, 255), 4, 4, 4, PieceTSquares);

//////////////////////////////////////////////////////////////////////
// Pieces
//////////////////////////////////////////////////////////////////////
Piece * pieces[] =
{
	&PieceO,
	&PieceI,
	&PieceS,
	&PieceZ,
	&PieceL,
	&PieceJ,
	&PieceT,
};

const float FACTOR = Field::PIECES_LENGTH * 1000.0f;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Field::Field()
{
	ZeroMemory(grid, sizeof (grid));

	piece = NULL;

	pieceTop = 0;
	pieceLeft = 0;

	lineCount = 0;

	gameOver = true;
}

void Field::NewGame()
{
	ZeroMemory(grid, sizeof (grid));

	piece = NULL;

	pieceTop = 0;
	pieceLeft = 0;

	lineCount = 0;

	gameOver = false;

	MakeNewPiece();
}

bool Field::CheckFree(int col, int row)
{
	if (col < 0 || col >= GRID_COLS)
	{
		return false;
	}
	if (row < 0 || row >= GRID_ROWS)
	{
		return false;
	}
	return grid[row][col] == NULL;
}

bool Field::CanOccupyPosition(int verticalOffset)
{
	Array2D colors;
	piece->GetColors(&colors);
	// for each col find whether there is
	// one free square verticalOffset rows below it:
	for (int col = piece->GetCols() - 1; col >= 0; --col)
	{
		int found = -1;
		for (int row = piece->GetRows() - 1; row >= 0; --row)
		{
			const int physicalRow = pieceTop + row;
			if (physicalRow >= 0)
			{
				if (colors.Get(row, col) != NULL)
				{
					if (!CheckFree(pieceLeft + col, physicalRow))
					{
						return false;
					}
					if (found == -1)
					{
						found = row;
					}
				}
			}
		}
		if (verticalOffset > 0 && found != -1)
		{
			if (
				!CheckFree(pieceLeft + col,
				pieceTop + found + verticalOffset)
			)
			{
				return false;
			}
		}
	}
	return true;
}

bool Field::FindOK(int col)
{
	Array2D colors;
	piece->GetColors(&colors);
	const int pieceLeftPlusCol = pieceLeft + col;
	{
		for (int row = piece->GetRows() - 1; row >= 0; --row)
		{
			const int physicalRow = pieceTop + row;
			if (
				(physicalRow >= 0) &&
				(colors.Get(row, col) != NULL) &&
				(!CheckFree(pieceLeftPlusCol, physicalRow))
			)
			{
				return false;
			}
		}
	}
	return true;
}

int FindCol(Piece * piece, int first, int afterLast, int step)
{
	Array2D colors;
	piece->GetColors(&colors);
	int rows = piece->GetRows();
	for (int col = first; col != afterLast; col += step)
	{
		for (int row = 0; row < rows; ++row)
		{
			if (colors.Get(row, col) != NULL)
			{
				return col;
			}
		}
	}
	return -1;
}

bool Field::Rotate()
{
	int prevPieceLeft = pieceLeft;
	piece->Rotate();
	bool succeeded = CanOccupyPosition(0);
	bool movePlus = true;
	bool moveMinus = true;
	const int cols = piece->GetCols();
	const int lastCol = FindCol(piece, cols - 1, -1, -1);
	const int firstCol = FindCol(piece, 0, cols, 1);
	const int realCols = lastCol - firstCol + 1;

	if (!succeeded)
	{
		for (int i = 0; i <= realCols ; i++)
		{
			if (movePlus)
			{
				pieceLeft = prevPieceLeft + i;
				succeeded = CanOccupyPosition(0);
				if (succeeded)
				{
					break;
				}
				else
				{
					if (!FindOK(lastCol))
					{
						movePlus = false;
					}
				}
			}
			if (moveMinus)
			{
				pieceLeft = prevPieceLeft - i;
				succeeded = CanOccupyPosition(0);
				if (succeeded)
				{
					break;
				}
				else
				{
					if (!FindOK(firstCol))
					{
						moveMinus = false;
					}
				}
			}
			if ((!movePlus) && (!moveMinus))
			{
				succeeded = false;
				break;
			}
		}
	}
	if (!succeeded)
	{
		piece->Unrotate();
		pieceLeft = prevPieceLeft;
		return false;
	}
	return true;
}

bool Field::CanMoveLeft()
{
	--pieceLeft;
	bool result = CanOccupyPosition(0);
	++pieceLeft;
	return result;
}

bool Field::CanMoveRight()
{
	++pieceLeft;
	bool result = CanOccupyPosition(0);
	--pieceLeft;
	return result;
}

bool Field::MoveLeft()
{
	if (CanMoveLeft())
	{
		--pieceLeft;
		return true;
	}
	return false;
}

bool Field::MoveRight()
{
	if (CanMoveRight())
	{
		++pieceLeft;
		return true;
	}
	return false;
}

bool Field::Drop()
{
	bool result = false;
	// Used to be : while (MoveDownward());
	// but this prevented the user
	// from moving the piece just before
	// it got settled.
	bool first = true;
	while (CanMoveDownward())
	{
		if (first)
		{
			result = true;
			first = false;
		}
		++pieceTop;
	}
	return result;
}

bool Field::CanMoveDownward()
{
	return CanOccupyPosition(1);
}

bool Field::MoveDown()
{
	if (CanMoveDownward())
	{
		++pieceTop;
		return true;
	}
	return false;
}

bool Field::MoveDownward()
{
	bool result = true;
	if (CanMoveDownward())
	{
		++pieceTop;
	}
	else
	{
		// copy it:
		CopyCurrentToGrid();
		// now, look for filled lines:
		for (int row = 0; row < GRID_ROWS; ++row)
		{
			if (RowFilled(row))
			{
				RemoveRow(row);
			}
		}
		// now, make a new piece:
		MakeNewPiece();
		result = false;
	}
	return result;
}

void Field::CopyCurrentToGrid()
{
	Array2D colors;
	piece->GetColors(&colors);
	// for each col find whether there is
	// one free square below it:
	for (int col = piece->GetCols() - 1; col >= 0; --col)
	{
		for (int row = piece->GetRows() - 1; row >= 0; --row)
		{
			int left = pieceLeft + col;
			int top = pieceTop + row;
			if (left < GRID_COLS && left >= 0 && top < GRID_ROWS && top >= 0)
			{
				if (colors.Get(row, col) != NULL)
				{
					grid[top][left] =
						colors.Get(row, col);
				}
			}
		}
	}
}

void Field::MakeNewPiece()
{
	int random = ((int)(MathRandom() * FACTOR)) % PIECES_LENGTH;
	piece = pieces[random];
	piece->Restore();
	pieceTop = -1;
	pieceLeft = (GRID_COLS - piece->GetCols()) / 2;

	if (!CanOccupyPosition(0))
	{
		gameOver = true;
		do
		{
			--pieceTop;
		}
		while (!CanOccupyPosition(0));
		CopyCurrentToGrid();
	}
}

bool Field::RowFilled(int row)
{
	for (int col = 0; col < GRID_COLS; ++col)
	{
		if (grid[row][col] == NULL)
		{
			return false;
		}
	}
	return true;
}

void Field::CopyRow(int from, int to)
{
	for (int col = 0; col < GRID_COLS; ++col)
	{
		grid[to][col] = grid[from][col];
	}
}

void Field::RemoveRow(int rowToRemove)
{
	// fix all lines except for the first one:
	for (int row = rowToRemove - 1; row >= 0; --row)
	{
		CopyRow(row, row + 1);
	}
	// fix the first line:
	for (int col = 0; col < GRID_COLS; ++col)
	{
		grid[0][col] = NULL;
	}
	// keep the score:
	++lineCount;
}

Square * Field::GetSquare(
	Piece * piece,
	int row,
	int col,
	int pieceLeft,
	int pieceTop,
	int index,
	bool gameOver
)
{
		int cols = piece->GetCols(index);
		int rows = piece->GetRows(index);
		if (
			!gameOver &&
			col >= pieceLeft &&
			col < pieceLeft + cols &&
			row >= pieceTop &&
			row < pieceTop + rows
		)
		{
			Array2D colors;
			piece->GetColors(index, &colors);
			return colors.Get(row - pieceTop, col - pieceLeft);
		}
		return NULL;
}

Square * Field::GetSquare(int row, int col)
{
	if (piece == NULL) return NULL;
	Square * s = GetSquare(piece, row, col, pieceLeft, pieceTop, piece->GetIndex(), gameOver);
	if (s != NULL) return s;
	return grid[row][col];
}

bool Field::IsFalling(int row, int col)
{
	if (piece == NULL) return false;
	int index = piece->GetIndex();
	int cols = piece->GetCols(index);
	int rows = piece->GetRows(index);
	return (
		!gameOver &&
		col >= pieceLeft &&
		col < pieceLeft + cols &&
		row >= pieceTop &&
		row < pieceTop + rows &&
		grid[row][col] == NULL
	);
}

bool Field::IsGameOver()
{
	return gameOver;
}

void Field::SetGameOver()
{
	gameOver = true;
}

int Field::GetCols()
{
	return GRID_COLS;
}

int Field::GetRows()
{
	return GRID_ROWS;
}

int Field::GetLineCount()
{
	return lineCount;
}
