// Field.h: interface for the Field class.
//
//////////////////////////////////////////////////////////////////////

#pragma once

#include "Square.h"
#include "Piece.h"

class Field
{
public:
	enum
	{
		GRID_ROWS = 20,
		GRID_COLS = 10,
		PIECES_LENGTH = 7
	};

	Field();
	void NewGame();
	bool CheckFree(int col, int row);
	bool CanOccupyPosition(int verticalOffset);
	bool FindOK(int col);
	bool Rotate();
	bool CanMoveLeft();
	bool CanMoveRight();
	bool MoveLeft();
	bool MoveRight();
	bool Drop();
	bool CanMoveDownward();
	bool MoveDown();
	bool MoveDownward();
	void CopyCurrentToGrid();
	void MakeNewPiece();
	bool RowFilled(int row);
	void CopyRow(int from, int to);
	void RemoveRow(int rowToRemove);
	static Square * GetSquare(
		Piece * piece,
		int row,
		int col,
		int pieceLeft,
		int pieceTop,
		int index,
		bool gameOver
	);
	Square * GetSquare(int row, int col);
	bool IsFalling(int row, int col);
	bool IsGameOver();
	void SetGameOver();
	int GetCols();
	int GetRows();
	int GetLineCount();

private:
	Square * grid[GRID_ROWS][GRID_COLS];

	Piece * piece;

	int pieceTop;
	int pieceLeft;

	int lineCount;

	bool gameOver;

};
