// GradientChange.cpp: implementation of the GradientChange class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "GradientChange.h"
#include "Gradient.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

static inline double Pow2(double x)
{
	return x * x;
}

static double Dist(int x1, int y1, double x2, double y2)
{
	double result = sqrt(Pow2(x1 - x2) + Pow2(y1 - y2));
	if (result <= 1.0)
	{
		return 1.0;
	}
	return result;
}

GradientChange::GradientChange()
{
}

void GradientChange::InitPoints()
{
	int i;
	Randomize();

	for (i = 0; i < GradientChange::INITIAL_POINTS; i++)
	{
		m_points_x[i] = MathRandom();
#ifdef GRADIENT_STARS_UP
		m_points_y[i] = 1.0 - (MathRandom() / 2.0);
#else
		m_points_y[i] = MathRandom();
#endif
	}
}

void GradientChange::Init()
{
	double dist;
	double sumDist;
	int x, y, i;

	const int w = Gradient::I_WIDTH;
	const int h = Gradient::I_HEIGHT;

	double points_x[GradientChange::INITIAL_POINTS];
	double points_y[GradientChange::INITIAL_POINTS];

	{
		const double dw = (double)Gradient::I_WIDTH;
		const double dh = (double)Gradient::I_HEIGHT;

		for (i = 0; i < GradientChange::INITIAL_POINTS; i++)
		{
			points_x[i] = m_points_x[i] * dw;
			points_y[i] = m_points_y[i] * dh;
		}
	}

	for (y = 0; y < h; y++)
	{
		for (x = 0; x < w; x++)
		{
			sumDist = 0.0;
			for (i = 0; i < GradientChange::INITIAL_POINTS; i++)
			{
				dist = pow(
					Dist(x, y, points_x[i], points_y[i]), -0.98);
				m_distances[x][y][i] = dist;
				sumDist += dist;
			}
			sumDist = 1.0 / sumDist;
			for (i = 0; i < GradientChange::INITIAL_POINTS; i++)
			{
				m_distances[x][y][i] = pow(m_distances[x][y][i] * sumDist, 1.05);
			}
		}
	}

}
