// GradientStar.cpp: implementation of the GradientStar class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "GradientStar.h"
#include "Star.h"
#include "HLS.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

static int randColorDarkHue()
{
	return MaxRandom(HLSMAX);
}

static int randColorDarkLum()
{
	return MaxRandom(80) + 70;
}

static int randColorDarkSat()
{
	return MaxRandom(40) + 200;
}

static void cutHLS(double & x)
{
	while (x > HLSMAX)
	{
		x -= HLSMAX;
	}
	while (x < 0)
	{
		x += HLSMAX;
	}
}

GradientStar::GradientStar()
{
	Randomize();
	hue = randColorDarkHue();
	lum = randColorDarkLum();
	sat = randColorDarkSat();
	current = HLStoRGB(
		MathRound<WORD>(hue),
		MathRound<WORD>(lum),
		MathRound<WORD>(sat));
	steps = 0;
	DoAdvancePrepare();
	int stepsToMake = MathRound<int>(
		MathRandom() * (steps - 50)) + 30;
	steps -= stepsToMake;
	double dSteps = double(stepsToMake);
	hue += stepHue * dSteps;
	lum += stepLum * dSteps;
	sat += stepSat * dSteps;
	cutHLS(hue);
	DoAdvance();
}

void GradientStar::Advance()
{
	DoAdvancePrepare();
	DoAdvance();
}

void GradientStar::DoAdvancePrepare()
{
	if (steps == 0)
	{
		steps = MathRound<int>(MathRandom() * 31) + 191;
		int targetHue = randColorDarkHue();
		int targetLum = randColorDarkLum();
		int targetSat = randColorDarkSat();
		double dSteps = double(steps);
		if (MathRandom() >= 0.5)
		{
			stepHue = (double(targetHue) - hue) / dSteps;
		}
		else
		{
			stepHue = (double(targetHue) - (hue + HLSMAX)) / dSteps;
		}
		stepLum = (double(targetLum) - lum) / dSteps;
		stepSat = (double(targetSat) - sat) / dSteps;
	}
}

void GradientStar::DoAdvance()
{
	--steps;
	hue += stepHue;
	lum += stepLum;
	sat += stepSat;
	cutHLS(hue);
	current = HLStoRGB(
		MathRound<WORD>(hue),
		MathRound<WORD>(lum),
		MathRound<WORD>(sat));
}

COLORREF GradientStar::GetColor()
{
	return current;
}
