// Piece.cpp: implementation of the Piece class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Piece.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
extern void ArmInitializeAll();
Piece::Piece(COLORREF color, int a, int b, int c, int * data):
allColorModes(a, b, c, squaresPtr)
{
	this->index = 0;
	this->prev = 0;

	this->color = color;
	this->pale = ModPale(color);
	this->dark = ModDark(color);

	ArmInitializeAll();
	this->colorBrush = PTR_EXTERNAL(CreateSolidBrush)(color);
	this->palePen = PTR_EXTERNAL(CreatePen)(PS_SOLID, 1, pale);
	this->darkPen = PTR_EXTERNAL(CreatePen)(PS_SOLID, 1, dark);

	int size = a * b * c;
	int i;
	for (i = 0; i < size; i++)
	{
		if (data[i] != -1)
		{
			squares[i].Set(data[i], this);
			squaresPtr[i] = &squares[i];
		}
		else
		{
			squares[i].Set(0, NULL);
			squaresPtr[i] = NULL;
		}
	}
}

Piece::~Piece()
{
	PTR_EXTERNAL(DeleteObject)(colorBrush);
	PTR_EXTERNAL(DeleteObject)(palePen);
	PTR_EXTERNAL(DeleteObject)(darkPen);
}

void Piece::GetColors(Array2D * arr)
{
	GetColors(index, arr);
}

void Piece::GetColors(int i, Array2D * arr)
{
	arr->Set(&allColorModes, i);
}

int Piece::GetCols()
{
	return GetCols(index);
}

int Piece::GetRows()
{
	return GetRows(index);
}

int Piece::GetCols(int i)
{
	return allColorModes.GetLength(i, 0);
}

int Piece::GetRows(int i)
{
	return allColorModes.GetLength(i);
}

void Piece::Rotate()
{
	prev = index;
	index = (index + 1) % allColorModes.GetLength();
}

void Piece::Unrotate()
{
	index = prev;
}

void Piece::Restore()
{
	index = 0;
}

int Piece::GetIndex()
{
	return index;
}
