// Square.cpp: implementation of the Square class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Square.h"
#include "Piece.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Square::Square()
{
	// Do Nothing
}

void Square::Set(int inOrientation, Piece * inPiece)
{
	this->orientation = inOrientation;
	this->piece = inPiece;
}

bool Square::Is(int dir)
{
	return (orientation & dir) != 0;
}

void Square::DrawLine(
	int dir,
	int x1, int y1, int x2, int y2,
	bool vert
)
{
	HPEN oldPen;

	if (Is(dir))
	{
		oldPen = (HPEN)PTR_EXTERNAL(SelectObject)(hDCGlobalBackBuffer, piece->palePen);
		if (vert)
		{
			DrawLineOffByOne(x1, y1, x1, y2 + 1);
		}
		else
		{
			DrawLineOffByOne(x1, y1, x2 + 1, y1);
		}
		PTR_EXTERNAL(SelectObject)(hDCGlobalBackBuffer, oldPen);
	}
	else
	{
		int ox1 = x1;
		int oy1 = y1;
		int ox2 = x2;
		int oy2 = y2;
		if (dir == D || dir == L)
		{
			oldPen = (HPEN)PTR_EXTERNAL(SelectObject)(hDCGlobalBackBuffer, piece->darkPen);
			if (vert)
			{
				y1 += BEG;
				y2 -= BEG;
				while (y1 < y2)
				{
					DrawLineOffByOne(x1, y1, x1, y1 + 2 + 1);
					y1 += 4;
				}
			}
			else
			{
				x1 += BEG;
				x2 -= BEG;
				while (x1 < x2)
				{
					DrawLineOffByOne(x1, y1, x1 + 2 + 1, y1);
					x1 += 4;
				}
			}
			PTR_EXTERNAL(SelectObject)(hDCGlobalBackBuffer, oldPen);
		}
		SetColor(piece->pale);
		PutPixel(ox1, oy1);
		PutPixel(ox2, oy2);
	}
}

void Square::PaintRectangle(
	int x,
	int y,
	int width,
	int height
)
{
	RECT rect;

	rect.left = x;
	rect.top = y;
	rect.right = x + width;
	rect.bottom = y + height;

	FillRect(hDCGlobalBackBuffer, &rect, piece->colorBrush);

	--rect.right;
	--rect.bottom;

	DrawLine(LEFT, x, y, x, rect.bottom, true);
	DrawLine(RIGHT, rect.right, y, rect.right, rect.bottom, true);
	DrawLine(UP, x, y, rect.right, y, false);
	DrawLine(DOWN, x, rect.bottom, rect.right, rect.bottom, false);
}
