// Star.cpp: implementation of the Star class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Star.h"
#include "Stage.h"
#include "StarColors.h"
#include "HLS.h"

#define STATIC_STARS (500)

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Star::Star():
colorStage(NUMBER_OF_STAGES - 1),
brightnessStage(SECOND_STAGE - 1)
{
	x = 0;
	iy = 0;
	yd = 0;
	speed = 0;

	Randomize();

	StarColors::RandomInterpolation(colors);

	next = NULL;
}

static int trim(int x, int max)
{
	if (x < 0) return 0;
	if (x > max) return max;
	return x;
}

static int randColor()
{
	return trim(((int)(MathRandom() * (255 - Star::COLOR_MIN))) + Star::COLOR_MIN, 255);
}

COLORREF Star::RandColorLight()
{
	return RGB(
		randColor(),
		randColor(),
		randColor()
	);
}

static bool firstTime = true;

static Star * buffer = NULL;
static Star * buffer2 = NULL;

static void MakeStars();

Star * Star::GetStar(int maxX, int maxY, Star * next)
{
	if (firstTime)
	{
		MakeStars();
		firstTime = false;
	}
	Star * s;
	if (buffer == NULL)
	{
		Star * t = buffer;
		buffer = buffer2;
		buffer2 = t;
	}
	if (buffer != NULL)
	{
		s = buffer;
		buffer = s->next;
	}
	else
	{
		return next;
	}
	s->speed = (float)MathRandom() * 5;
	if (s->speed < 0.2f) s->speed = 0.2f;
	s->x = trim(MathRound<int>((float)MathRandom() * maxX), maxX);
	s->yd = (float)maxY + s->speed - 0.01f + 2.0f;
	s->iy = MathRound<int>(s->yd);
	s->next = next;
	return s;
}

Star * Star::PutStar(Star *s)
{
	Star * prevNext = s->next;
	s->next = buffer2;
	buffer2 = s;
	return prevNext;
}

static Star staticStars[STATIC_STARS];

static void MakeStars()
{
	int i;
	for (i = 0; i < STATIC_STARS; i++)
	{
		Star::PutStar(&staticStars[i]);
	}
}

void Star::Advance()
{
	yd -= speed;
	iy = MathRound<int>(yd);
	colorStage.Advance();
	brightnessStage.Advance();
}

void Star::RemoveNext()
{
	if (next != NULL)
	{
		next = Star::PutStar(next);
	}
}

void Star::Paint()
{
	colors[colorStage.GetValue()].Paint(
		brightnessStage.GetValue() / 10, x, iy);
}

COLORREF Star::GetColor()
{
	return colors[colorStage.GetValue()].centerColor;
}
