// StarColors.cpp: implementation of the StarColors class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "StarColors.h"
#include "Star.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

StarColors::StarColors()
{
}

COLORREF StarColors::Init()
{
	COLORREF color;

	color = Star::RandColorLight();
	Init(color);
	return color;
}

void StarColors::Init(COLORREF color)
{
	centerColor = color;
	downColor = ModDark(color);
	upColor = ModPale(color);
	brightColor = ModPale(upColor);
}

void StarColors::RandomInterpolation(StarColors * result)
{
	int n = Star::NUMBER_OF_STAGES - 1;

	COLORREF color0 = result[0].Init();
	COLORREF colorN = result[n].Init();
	// make interpolation:
	int r1 = GetRValue(color0);
	int g1 = GetGValue(color0);
	int b1 = GetBValue(color0);
	int r2 = GetRValue(colorN);
	int g2 = GetGValue(colorN);
	int b2 = GetBValue(colorN);
	double rdiff = r2 - r1;
	double gdiff = g2 - g1;
	double bdiff = b2 - b1;
	double factor = ((double)1.0) / ((double)Star::NUMBER_OF_STAGES);
	for (int i = 1; i < n; i++)
	{
		int r = r1 + (MathRound<int>(factor * i * rdiff));
		int g = g1 + (MathRound<int>(factor * i * gdiff));
		int b = b1 + (MathRound<int>(factor * i * bdiff));
		result[i].Init(RGB(r, g, b));
	}
}

void StarColors::Paint(int brightness, int x, int iy)
{
	if (brightness == 4)
	{
		if (
			x >= ClipX1 - 2 && x <= ClipX2 + 2 &&
			iy >= ClipY1 - 2 && iy <= ClipY2 + 2
		)
		{
			SetColor(brightColor);
			PutPixel(x, iy);
			SetColor(upColor);
			PutPixel(x - 1, iy);
			PutPixel(x + 1, iy);
			PutPixel(x, iy - 1);
			PutPixel(x, iy + 1);
			SetColor(centerColor);
			PutPixel(x - 1, iy - 1);
			PutPixel(x + 1, iy + 1);
			PutPixel(x - 1, iy + 1);
			PutPixel(x + 1, iy - 1);
			SetColor(downColor);
			PutPixel(x - 2, iy);
			PutPixel(x + 2, iy);
			PutPixel(x, iy - 2);
			PutPixel(x, iy + 2);
		}
	}
	else if (brightness == 3)
	{
		if (
			x >= ClipX1 - 2 && x <= ClipX2 + 2 &&
			iy >= ClipY1 - 2 && iy <= ClipY2 + 2
		)
		{
			SetColor(upColor);
			PutPixel(x, iy);
			PutPixel(x - 1, iy);
			PutPixel(x + 1, iy);
			PutPixel(x, iy - 1);
			PutPixel(x, iy + 1);
			SetColor(centerColor);
			PutPixel(x - 1, iy - 1);
			PutPixel(x + 1, iy + 1);
			PutPixel(x - 1, iy + 1);
			PutPixel(x + 1, iy - 1);
			SetColor(downColor);
			PutPixel(x - 2, iy);
			PutPixel(x + 2, iy);
			PutPixel(x, iy - 2);
			PutPixel(x, iy + 2);
		}
	}
	else if (brightness == 2)
	{
		if (
			x >= ClipX1 - 2 && x <= ClipX2 + 2 &&
			iy >= ClipY1 - 2 && iy <= ClipY2 + 2
		)
		{
			SetColor(upColor);
			PutPixel(x, iy);
			SetColor(centerColor);
			PutPixel(x - 1, iy);
			PutPixel(x + 1, iy);
			PutPixel(x, iy - 1);
			PutPixel(x, iy + 1);
			PutPixel(x - 1, iy - 1);
			PutPixel(x + 1, iy + 1);
			PutPixel(x - 1, iy + 1);
			PutPixel(x + 1, iy - 1);
			SetColor(downColor);
			PutPixel(x - 2, iy);
			PutPixel(x + 2, iy);
			PutPixel(x, iy - 2);
			PutPixel(x, iy + 2);
		}
	}
	else if (brightness == 1)
	{
		if (
			x >= ClipX1 - 2 && x <= ClipX2 + 2 &&
			iy >= ClipY1 - 2 && iy <= ClipY2 + 2
		)
		{
			SetColor(upColor);
			PutPixel(x, iy);
			SetColor(centerColor);
			PutPixel(x - 1, iy);
			PutPixel(x + 1, iy);
			PutPixel(x, iy - 1);
			PutPixel(x, iy + 1);
			SetColor(downColor);
			PutPixel(x - 1, iy - 1);
			PutPixel(x + 1, iy + 1);
			PutPixel(x - 1, iy + 1);
			PutPixel(x + 1, iy - 1);
			PutPixel(x - 2, iy);
			PutPixel(x + 2, iy);
			PutPixel(x, iy - 2);
			PutPixel(x, iy + 2);
		}
	}
	else if (brightness == 0)
	{
		if (
			x >= ClipX1 - 1 && x <= ClipX2 + 1 &&
			iy >= ClipY1 - 1 && iy <= ClipY2 + 1
		)
		{
			SetColor(upColor);
			PutPixel(x, iy);
			SetColor(centerColor);
			PutPixel(x - 1, iy);
			PutPixel(x + 1, iy);
			PutPixel(x, iy - 1);
			PutPixel(x, iy + 1);
			SetColor(downColor);
			PutPixel(x - 1, iy - 1);
			PutPixel(x + 1, iy + 1);
			PutPixel(x - 1, iy + 1);
			PutPixel(x + 1, iy - 1);
		}
	}
}
