#pragma once

#ifndef __GNUC__
#pragma warning( disable : 4820)
#pragma warning( disable : 4668)
#pragma warning( disable : 4710)
#pragma warning( disable : 4711)
#endif

////////////////////////////////////////
// Configuration Definitions
////////////////////////////////////////

// Use project options to configure UNICODE support.
// Unicode NOT supported for GCC.
// Standard build is *without* unicode, to support Win9x.

// Define this to get settings for slow machines
#define SLOW_MACHINE

// Define this to get high-quality, slow background
// #define TETRIS_BACKGROUND_HIGH_QUALITY

// Define this to stretch the background, if necessary
// If this is not defined, the window will not be sizable
// #define STRETCH

// Define this to place all the background
// lights in the upper part of the screen
// #define GRADIENT_STARS_UP

// Define this to beep at each game tick
// #define BEEP_MOVE

// Define this to have smooth falling
#define SMOOTH_FALLING

// Define this to have smooth left/right/down movement
//#define SMOOTH_ACCELERATE

#define WIDTH_FACTOR (6.28)
#define HEIGHT_FACTOR (6.28)

////////////////////////////////////////
// Other Definitions
////////////////////////////////////////

#ifdef SLOW_MACHINE
	#ifdef TETRIS_BACKGROUND_HIGH_QUALITY
		#undef TETRIS_BACKGROUND_HIGH_QUALITY
	#endif
	#ifndef STRETCH
		#define STRETCH
	#endif
	#ifdef SMOOTH_ACCELERATE
		#undef SMOOTH_ACCELERATE
	#endif
#endif

#define BIG_PIC_SIZE_X (270)
#define BIG_PIC_SIZE_Y (540)

#define INITIAL_WIDTH_0 (BIG_PIC_SIZE_X + 9)
#define INITIAL_HEIGHT_0 (BIG_PIC_SIZE_Y + 34)

#ifdef STRETCH
	#define INITIAL_WIDTH (INITIAL_WIDTH_0)
	#define INITIAL_HEIGHT (INITIAL_HEIGHT_0)
#else
	#define INITIAL_WIDTH (INITIAL_WIDTH_0 - 2)
	#define INITIAL_HEIGHT (INITIAL_HEIGHT_0 - 2)
#endif

#ifdef TETRIS_BACKGROUND_HIGH_QUALITY
	#ifndef WIDTH_FACTOR
		#define WIDTH_FACTOR (1)
	#endif
	#ifndef HEIGHT_FACTOR
		#define HEIGHT_FACTOR (1)
	#endif
#else
#ifdef SLOW_MACHINE
	#ifndef WIDTH_FACTOR
		#define WIDTH_FACTOR (5)
	#endif
	#ifndef HEIGHT_FACTOR
		#define HEIGHT_FACTOR (5)
	#endif
#else
	#ifndef WIDTH_FACTOR
		#define WIDTH_FACTOR (2.39)
	#endif
	#ifndef HEIGHT_FACTOR
		#define HEIGHT_FACTOR (2.39)
	#endif
#endif
#endif

#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers

// Use these for cross-compilation support for newer compilers
#define _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES (1)
#define _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES_COUNT (1)

#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include <windows.h>
#ifndef __GNUC__
#pragma warning(disable:4201)
#endif
#include <mmsystem.h>
#ifndef __GNUC__
#pragma warning(default:4201)
#endif
#include <wingdi.h>

// for the gradient:
#include <math.h>

extern "C"
{
	WINGDIAPI BOOL WINAPI AlphaBlend(HDC,int,int,int,int,HDC,int,int,int,int,BLENDFUNCTION);
}

void Randomize();

extern HDC hDCGlobalBackBuffer;
extern int ClipX1;
extern int ClipY1;
extern int ClipX2;
extern int ClipY2;

void SetColor(COLORREF color);
void PutPixel(int x, int y);
void DrawLineOffByOne(int x1, int y1, int x2, int y2);

double MathRandom();
template <typename T> T MathRound(double value)
{
	return (T)(value + 0.5);
}
int MaxRandom(int max);

COLORREF ModDark(COLORREF color);
COLORREF ModPale(COLORREF color);
COLORREF Mod(int r, int g, int b);

#define ELEMENTS(X) ((sizeof( (X) )) / (sizeof( (X)[0] )))

#ifndef tstrlen
	#ifdef UNICODE
		#define tstrlen wcslen
	#else
		#define tstrlen strlen
	#endif
#endif

#ifndef _sntprintf
	#ifdef UNICODE
		#define _sntprintf _snwprintf
	#else
		#define _sntprintf _snprintf
	#endif
#endif

#ifdef _M_ARM
#define TETRIS_ARM_SUPPORT (1)
#endif
#include "ArmSupport.h"
