// Timer.cpp: implementation of the Timer class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Timer.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// scale factor for converting a performancce counter reading into milliseconds:
static __int64 s_scaleFactor;

void InitTimerLibrary()
{
	__int64 counterFrequency;

	::QueryPerformanceFrequency((LARGE_INTEGER *)(&counterFrequency));

	// NOTE: counterFrequency will be zero for a machine that does not have
	// support for a high-resolution counter. This is only likely for very
	// old hardware but for a robust implementation you should handle this
	// case.

	s_scaleFactor = counterFrequency;
}

static inline __int64 GetTimeM()
{
	__int64 counterReading;

	::QueryPerformanceCounter((LARGE_INTEGER *)(&counterReading));

	return (counterReading * 1000) / s_scaleFactor;
}

static __int64 GetTime()
{
	return ::GetTimeM();
}

Timer::Timer(int millis)
{
	this->m_next = 0;
	this->m_millis = millis;
	this->m_millisToSkip = -(millis << 1);
}

void Timer::Start()
{
	this->m_next = ::GetTime() + this->m_millis;
}

void Timer::Proceed()
{
	__int64 prev = this->m_next;
	this->m_next += this->m_millis;
	this->Proceed(prev);
}

void Timer::Proceed(__int64 to)
{
	int sleepTime = (int)(to - ::GetTimeM());
	int millis = this->m_millis;
	if (sleepTime > 0 && sleepTime <= millis)
	{
		::Sleep((DWORD)sleepTime);
	}
	else if (sleepTime < m_millisToSkip || sleepTime > millis)
	{
		// Computer coming out of sleep:
		this->Start();
		::Sleep((DWORD)this->m_millis);
	}
}
