// Win32Tetris.cpp : Defines the entry point for the application.
//

#include "stdafx.h"
#include "Field.h"
#include "Timer.h"
#include "StarField.h"
#include "Square.h"
#include "Gradient.h"
#include "Mutex.h"
#include "MovementStage.h"

// Keyboard definitions
#define VK_OTHER_KP5 (0xC)
#define VK_A (0x41)
#define VK_D (0x44)
#define VK_I (0x49)
#define VK_J (0x4A)
#define VK_K (0x4B)
#define VK_L (0x4C)
#define VK_P (0x50)
#define VK_S (0x53)
#define VK_W (0x57)

// User message, used to finish window initialization.
#define WMU_FINISH (WM_USER + 3)

static Gradient gradient;
static GradientChange gradientChange;

// Global Variables:
static HWND hWnd;
static const TCHAR * const szTitle = TEXT("Tetris");		// The title bar text
static const TCHAR * const szWindowClass = TEXT("Tetris");	// Window class name

// Lock them in this order ONLY!
static Mutex lockBackground;
static Mutex lockGUI;
static Mutex lockMutual;

POINT size;

int ClipX1 = 0;
int ClipY1 = 0;
int ClipX2 = 0;
int ClipY2 = 0;

COLORREF PaintingColor = 0;

HDC hDCWindow = NULL;

HDC hDCGlobalBackBuffer = NULL;

static HBITMAP hBitmapGlobalBackBuffer = NULL;
static HBITMAP hOldBitmapGlobalBackBuffer = NULL;

#ifdef SMOOTH_ACCELERATE
static HDC hDCAlphaBuffer = NULL;
static HBITMAP hBitmapAlphaBuffer = NULL;
static HBITMAP hOldBitmapAlphaBuffer = NULL;
#endif

static HDC hDCBackground = NULL;
static HBITMAP hBitmapBackground = NULL;
static HBITMAP hOldBitmapBackground = NULL;

static HDC hDCBackgroundBB = NULL;
static HBITMAP hBitmapBackgroundBB = NULL;
static HBITMAP hOldBitmapBackgroundBB = NULL;

static HBRUSH black = NULL;
static HBRUSH white = NULL;

static HFONT textFont = NULL;
static HFONT oldFont = NULL;

static bool created = false;
static bool randomized = false;

// Game:
#define VERTICAL_TEXT_START (3)
#define HORIZONTAL_TEXT_START (5)
StarField starField;
Field field;
int oldLineCount = -1;
TCHAR lineCountBuf[20];
int timerIndex = 0;
int savedTimerIndex = 0;
bool paused = false;

// Foward declarations of functions included in this code module:
static ATOM RegisterMainWindowClass(HINSTANCE hInstance);
BOOL InitInstance(HINSTANCE, int);
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
static DWORD WINAPI TetrisTickProc(LPVOID lpParameter);

const int TIMER_INDEX = 14;
const int MOVEMENT_STAGE_MAX = TIMER_INDEX;
// counts downwards:
MovementStage movementStageLeft(MOVEMENT_STAGE_MAX);
MovementStage movementStageRight(MOVEMENT_STAGE_MAX);
MovementStage movementStageDownwards(MOVEMENT_STAGE_MAX);

#ifdef SMOOTH_ACCELERATE
BLENDFUNCTION blendFunction;
#endif

#ifdef TETRIS_ARM_SUPPORT

/////////////////////
// timeBeginPeriod
/////////////////////
DEFINE_EXTERNAL(timeBeginPeriod) = NULL;

/////////////////////
// timeEndPeriod
/////////////////////
DEFINE_EXTERNAL(timeEndPeriod) = NULL;

/////////////////////
// CreateSolidBrush
/////////////////////
DEFINE_EXTERNAL(CreateSolidBrush) = NULL;

/////////////////////
// CreatePen
/////////////////////
DEFINE_EXTERNAL(CreatePen) = NULL;

/////////////////////
// DeleteObject
/////////////////////
DEFINE_EXTERNAL(DeleteObject) = NULL;

/////////////////////
// SelectObject
/////////////////////
DEFINE_EXTERNAL(SelectObject) = NULL;

//////////////////////
// SetPixel
//////////////////////
DEFINE_EXTERNAL(SetPixel) = NULL;

///////////////////////
// MoveToEx
///////////////////////
DEFINE_EXTERNAL(MoveToEx) = NULL;

////////////////////////
// LineTo
////////////////////////
DEFINE_EXTERNAL(LineTo) = NULL;

////////////////////////
// TextOut
////////////////////////
DEFINE_EXTERNAL(TextOut) = NULL;

///////////////////////
// GetDeviceCaps
///////////////////////
DEFINE_EXTERNAL(GetDeviceCaps) = NULL;

////////////////////////
// CreateFont
////////////////////////
DEFINE_EXTERNAL(CreateFont) = NULL;

DEFINE_EXTERNAL(DeleteDC);
DEFINE_EXTERNAL(CreateCompatibleDC);
DEFINE_EXTERNAL(CreateCompatibleBitmap);
DEFINE_EXTERNAL(SetTextColor);
DEFINE_EXTERNAL(SetBkMode);
DEFINE_EXTERNAL(BitBlt);
DEFINE_EXTERNAL(SetStretchBltMode);
DEFINE_EXTERNAL(SetBrushOrgEx);
DEFINE_EXTERNAL(SetDIBits);
DEFINE_EXTERNAL(StretchBlt);

void ArmInitializeAll()
{
	BODY_EXTERNAL(timeBeginPeriod, timeBeginPeriod, winmm.dll);
	BODY_EXTERNAL(timeEndPeriod, timeEndPeriod, winmm.dll);
	BODY_EXTERNAL(CreateSolidBrush, CreateSolidBrush, gdi32.dll);
	BODY_EXTERNAL(CreatePen, CreatePen, gdi32.dll);
	BODY_EXTERNAL(DeleteObject, DeleteObject, gdi32.dll);
	BODY_EXTERNAL(SelectObject, SelectObject, gdi32.dll);
	BODY_EXTERNAL(SetPixel, SetPixel, gdi32.dll);
	BODY_EXTERNAL(MoveToEx, MoveToEx, gdi32.dll);
	BODY_EXTERNAL(LineTo, LineTo, gdi32.dll);
#ifdef UNICODE
	BODY_EXTERNAL(TextOut, TextOutW, gdi32.dll);
#else
	BODY_EXTERNAL(TextOut, TextOutA, gdi32.dll);
#endif
	BODY_EXTERNAL(GetDeviceCaps, GetDeviceCaps, gdi32.dll);
#ifdef UNICODE
	BODY_EXTERNAL(CreateFont, CreateFontW, gdi32.dll);
#else
	BODY_EXTERNAL(CreateFont, CreateFontA, gdi32.dll);
#endif
	BODY_EXTERNAL(DeleteDC, DeleteDC, gdi32.dll);
	BODY_EXTERNAL(CreateCompatibleDC, CreateCompatibleDC, gdi32.dll);
	BODY_EXTERNAL(CreateCompatibleBitmap, CreateCompatibleBitmap, gdi32.dll);
	BODY_EXTERNAL(SetTextColor, SetTextColor, gdi32.dll);
	BODY_EXTERNAL(SetBkMode, SetBkMode, gdi32.dll);
	BODY_EXTERNAL(BitBlt, BitBlt, gdi32.dll);
	BODY_EXTERNAL(SetStretchBltMode, SetStretchBltMode, gdi32.dll);
	BODY_EXTERNAL(SetBrushOrgEx, SetBrushOrgEx, gdi32.dll);
	BODY_EXTERNAL(SetDIBits, SetDIBits, gdi32.dll);
	BODY_EXTERNAL(StretchBlt, StretchBlt, gdi32.dll);
}

#else

#define BODY_EXTERNAL(FunctionName, LoadName, Library) \
	if (p_ ## FunctionName == NULL) \
	{ \
		HMODULE hModule = LoadLibrary(_T( #Library ) ); \
		if (hModule != NULL) \
		{ \
			p_ ## FunctionName = (t_ ## FunctionName)(void *)GetProcAddress(hModule, #LoadName ); \
		} \
	}

/////////////////////
// timeBeginPeriod
/////////////////////
typedef MMRESULT (WINAPI *t_timeBeginPeriod)( UINT uPeriod);
t_timeBeginPeriod p_timeBeginPeriod = NULL;

/////////////////////
// timeEndPeriod
/////////////////////
typedef MMRESULT (WINAPI *t_timeEndPeriod)( UINT uPeriod);
t_timeEndPeriod p_timeEndPeriod = NULL;

void ArmInitializeAll()
{
	BODY_EXTERNAL(timeBeginPeriod, timeBeginPeriod, winmm.dll);
	BODY_EXTERNAL(timeEndPeriod, timeEndPeriod, winmm.dll);
}

#endif

int MathMin(int a, int b)
{
	if (a < b) return a;
	return b;
}

int MathMax(int a, int b)
{
	if (a > b) return a;
	return b;
}

static int Mod(int i)
{
	return MathMin(MathRound<int>((float)i * 0.85), 255);
}

static int ModRed(int i)
{
	return Mod(i + 30);
}

COLORREF Mod(int r, int g, int b)
{
	return RGB(
		ModRed(r),
		Mod(g),
		Mod(b)
	);
}

static int ModUp(int i)
{
	i += 25;
	return MathMin(MathRound<int>((float)i * 1.2), 255);
}

COLORREF ModPale(COLORREF c)
{
	return RGB(
		ModUp(GetRValue(c)),
		ModUp(GetGValue(c)),
		ModUp(GetBValue(c))
	);
}

static int ModDown(int i)
{
	i -= 30;
	return MathMax(MathRound<int>((float)i * 0.85), 0);
}

COLORREF ModDark(COLORREF c)
{
	return RGB(
		ModDown(GetRValue(c)),
		ModDown(GetGValue(c)),
		ModDown(GetBValue(c))
	);
}

void Randomize()
{
	if (!randomized)
	{
		LARGE_INTEGER counterReading;

		QueryPerformanceCounter(&counterReading);

		srand(counterReading.LowPart ^ counterReading.HighPart);

		randomized = true;
	}
}

double MathRandom()
{
	return double(rand()) / double(RAND_MAX);
}

int MaxRandom(int max)
{
	return MathRound<int>(MathRandom() * max);
}

static void GetSize()
{
	RECT rect;
	GetClientRect(hWnd, &rect);
	size.x = rect.right;
	size.y = rect.bottom;
}

void SetColor(COLORREF color)
{
	PaintingColor = color;
}

void SetClip(int x1, int y1, int x2, int y2)
{
	ClipX1 = x1;
	ClipY1 = y1;
	ClipX2 = x2;
	ClipY2 = y2;
}

void PutPixel(int x, int y)
{
	if (
		x >= ClipX1 && x <= ClipX2 &&
		y >= ClipY1 && y <= ClipY2
	)
	{
		PTR_EXTERNAL(SetPixel)(hDCGlobalBackBuffer, x, y, PaintingColor);
	}
}

void DrawLineOffByOne(int x1, int y1, int x2, int y2)
{
	PTR_EXTERNAL(MoveToEx)(hDCGlobalBackBuffer, x1, y1, NULL);
	PTR_EXTERNAL(LineTo)(hDCGlobalBackBuffer, x2, y2);
}

void DrawString(const TCHAR * const text, int x, int y)
{
	PTR_EXTERNAL(TextOut)(hDCGlobalBackBuffer, x, y, text, int(tstrlen(text)));
}

int APIENTRY WinMain(
	HINSTANCE hInstance,
	HINSTANCE /*hPrevInstance*/,
	LPSTR /*lpCmdLine*/,
	int nCmdShow
)
{
	MSG msg;
	BOOL bRet;
	DWORD threadID;
	HANDLE threadHandle;

#ifdef SMOOTH_ACCELERATE
	blendFunction.BlendOp = AC_SRC_OVER;
	blendFunction.BlendFlags = 0;
	blendFunction.SourceConstantAlpha = 127;
	blendFunction.AlphaFormat = 0;
#endif

	ArmInitializeAll();
	{
		Synchronized sync0(lockBackground);
		Synchronized sync1(lockGUI);

		Randomize();

		p_timeBeginPeriod(1);
		InitTimerLibrary();

		// Initialize global strings
		RegisterMainWindowClass(hInstance);

		// Perform application initialization:
		if (!InitInstance(hInstance, nCmdShow))
		{
			p_timeEndPeriod(1);
			ExitProcess(0);
		}

		GetSize();

		gradient.Init();
		CopyMemory(&gradientChange, &gradient.m_change, sizeof (GradientChange));

		threadHandle = CreateThread(
			NULL, // LPSECURITY_ATTRIBUTES lpThreadAttributes, // SD
			0, // SIZE_T dwStackSize, // initial stack size
			(LPTHREAD_START_ROUTINE)TetrisTickProc, // LPTHREAD_START_ROUTINE lpStartAddress, // thread function
			0, // LPVOID lpParameter, // thread argument
			0, // DWORD dwCreationFlags, // creation option
			&threadID // LPDWORD lpThreadId // thread identifier
		);
		SetThreadPriority(threadHandle, THREAD_PRIORITY_HIGHEST);
	}

	// Main message loop:
	for (;;)
	{
		// Request ownership of the critical section.
		bRet = GetMessage(&msg, NULL, 0, 0 );
		if (bRet == FALSE)
		{
			break;
		}
		if (bRet == -1)
		{
			// handle the error and possibly exit
			break;
		}
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	p_timeEndPeriod(1);
	ExitProcess(0);
	// Unreachable code:
	#ifndef _MSC_VER
	#define _MSC_VER 0
	#endif
    #if (_MSC_VER != 1900)
    #if (_MSC_VER != 1800)
    #if (_MSC_VER != 1600)
    #if (_MSC_VER != 1500)
	return 0;
    #endif
	#endif
    #endif
	#endif
}

//
//  FUNCTION: RegisterMainWindowClass()
//
//  PURPOSE: Registers the window class.
//
//  COMMENTS:
//
//    This function and its usage is only necessary if you want this code
//    to be compatible with Win32 systems prior to the 'RegisterClassEx'
//    function that was added to Windows 95. It is important to call this function
//    so that the application will get 'well formed' small icons associated
//    with it.
//
static ATOM RegisterMainWindowClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;

	ZeroMemory(&wcex, sizeof (WNDCLASSEX));

	wcex.cbSize = sizeof (WNDCLASSEX);

	wcex.style = CS_OWNDC;
	wcex.lpfnWndProc = (WNDPROC)WndProc;
	wcex.hInstance = hInstance;
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.lpszClassName = szWindowClass;

	return RegisterClassEx(&wcex);
}

//
//   FUNCTION: InitInstance(HANDLE, int)
//
//   PURPOSE: Saves instance handle and creates main window
//
//   COMMENTS:
//
//        In this function, we save the instance handle in a global variable and
//        create and display the main program window.
//
BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
	//   hInst = hInstance; // Store instance handle in our global variable
#ifdef STRETCH
	const int style = WS_OVERLAPPEDWINDOW;
#else
	const int style = WS_OVERLAPPED | WS_MINIMIZEBOX | WS_CAPTION | WS_SYSMENU;
#endif

	hWnd = CreateWindow(szWindowClass, szTitle, style,
		CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

	if (!hWnd)
	{
		return FALSE;
	}

	hDCWindow = GetDC(hWnd);
	int scrWidth = PTR_EXTERNAL(GetDeviceCaps)(hDCWindow, HORZRES);

	SetWindowPos(
		hWnd,                           // handle to window
		HWND_TOP,                       // placement-order handle
		(scrWidth - INITIAL_WIDTH) / 2, // horizontal position
#ifndef STRETCH
		1 +
#endif
		5,                              // vertical position
		INITIAL_WIDTH,                  // width
		INITIAL_HEIGHT,                 // height
		SWP_DEFERERASE |                // window-positioning options
		SWP_DRAWFRAME |
		SWP_FRAMECHANGED |
		SWP_NOCOPYBITS |
		SWP_NOREDRAW |
		SWP_NOSENDCHANGING
	);

	GetSize();

	const POINT difference =
	{
		BIG_PIC_SIZE_X - size.x,
		BIG_PIC_SIZE_Y - size.y
	};

	const int newSizeX = INITIAL_WIDTH + difference.x;
	const int newSizeY = INITIAL_HEIGHT + difference.y;

	SetWindowPos(
		hWnd,                           // handle to window
		HWND_TOP,                       // placement-order handle
		(scrWidth - newSizeX) / 2,      // horizontal position
#ifndef STRETCH
		1 +
#endif
		5,                              // vertical position
		newSizeX,                       // width
		newSizeY,                       // height
		SWP_DEFERERASE |                // window-positioning options
		SWP_DRAWFRAME |
		SWP_FRAMECHANGED |
		SWP_NOCOPYBITS |
		SWP_NOREDRAW |
		SWP_NOSENDCHANGING |
		SWP_SHOWWINDOW
	);

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);
	SendMessage(
		hWnd,
		WM_NCPAINT,
		1,
		0
	);
	return TRUE;
}

static void BlitBackground()
{
#ifdef STRETCH
	if (size.x != Gradient::PIC_SIZE_X || size.y != Gradient::PIC_SIZE_Y)
	{
		PTR_EXTERNAL(StretchBlt)(
			hDCBackground,    // handle to destination DC
			0,          // x-coord of destination upper-left corner
			0,          // y-coord of destination upper-left corner
			size.x,     // width of destination rectangle
			size.y,     // height of destination rectangle
			hDCBackgroundBB,   // handle to source DC
			0,          // x-coord of source upper-left corner
			0,          // y-coord of source upper-left corner
			Gradient::PIC_SIZE_X, // width of source rectangle
			Gradient::PIC_SIZE_Y, // height of source rectangle
			SRCCOPY     // raster operation code
		);
	}
	else
	{
#endif
		PTR_EXTERNAL(BitBlt)(hDCBackground, 0, 0, size.x, size.y, hDCBackgroundBB, 0, 0, SRCCOPY);
#ifdef STRETCH
	}
#endif
}

static void FixStretching(HDC hDC)
{
	PTR_EXTERNAL(SetStretchBltMode)(hDC, HALFTONE);
	PTR_EXTERNAL(SetBrushOrgEx)(hDC, 0, 0, NULL);
}

static void BlitBackgroundBB()
{
	PTR_EXTERNAL(SetDIBits)(
		hDCBackgroundBB,               // handle to DC
		hBitmapBackgroundBB,               // handle to hBitmapGlobalBackBuffer
		0,                      // starting scan line
		Gradient::PIC_SIZE_Y,   // number of scan lines
		(const void *)gradient.m_bits,        // array of hBitmapGlobalBackBuffer bits
		(const BITMAPINFO *)&gradient.m_bitmapInfo, // hBitmapGlobalBackBuffer data
		DIB_RGB_COLORS          // type of color indexes to use
	);

}

static void DeleteMemDC()
{
	if (hDCGlobalBackBuffer != NULL)
	{
		PTR_EXTERNAL(SelectObject)(hDCGlobalBackBuffer, hOldBitmapGlobalBackBuffer);
		PTR_EXTERNAL(SelectObject)(hDCGlobalBackBuffer, oldFont);
		PTR_EXTERNAL(DeleteObject)(hBitmapGlobalBackBuffer);
		PTR_EXTERNAL(DeleteDC)(hDCGlobalBackBuffer);
		hDCGlobalBackBuffer = NULL;

#ifdef SMOOTH_ACCELERATE
		PTR_EXTERNAL(SelectObject)(hDCAlphaBuffer, hOldBitmapAlphaBuffer);
		PTR_EXTERNAL(DeleteObject)(hBitmapAlphaBuffer);
		PTR_EXTERNAL(DeleteDC)(hDCAlphaBuffer);
		hDCAlphaBuffer = NULL;
#endif

		PTR_EXTERNAL(SelectObject)(hDCBackground, hOldBitmapBackground);
		PTR_EXTERNAL(DeleteObject)(hBitmapBackground);
		PTR_EXTERNAL(DeleteDC)(hDCBackground);
		hDCBackground = NULL;

	}
}

static void DeleteBackgroundBB()
{
	if (hDCBackgroundBB != NULL)
	{
		PTR_EXTERNAL(SelectObject)(hDCBackgroundBB, hOldBitmapBackgroundBB);
		PTR_EXTERNAL(DeleteObject)(hBitmapBackgroundBB);
		PTR_EXTERNAL(DeleteDC)(hDCBackgroundBB);
		hDCBackgroundBB = NULL;
	}
}

static void CreateMemDC()
{
	if (hDCGlobalBackBuffer != NULL)
	{
		DeleteMemDC();
	}
	GetSize();
	hDCGlobalBackBuffer = PTR_EXTERNAL(CreateCompatibleDC)(hDCWindow);
	if (hDCGlobalBackBuffer == NULL)
	{
		p_timeEndPeriod(1);
		ExitProcess(1);
	}
	hBitmapGlobalBackBuffer = PTR_EXTERNAL(CreateCompatibleBitmap)(hDCWindow, size.x, size.y);
	if (hBitmapGlobalBackBuffer == NULL)
	{
		p_timeEndPeriod(1);
		ExitProcess(1);
	}
	hOldBitmapGlobalBackBuffer = (HBITMAP)PTR_EXTERNAL(SelectObject)(
		hDCGlobalBackBuffer, hBitmapGlobalBackBuffer);
	oldFont = (HFONT)PTR_EXTERNAL(SelectObject)(hDCGlobalBackBuffer, textFont);
	PTR_EXTERNAL(SetTextColor)(hDCGlobalBackBuffer, RGB(255, 255, 255));
	PTR_EXTERNAL(SetBkMode)(hDCGlobalBackBuffer, TRANSPARENT);

#ifdef SMOOTH_ACCELERATE
	hDCAlphaBuffer = PTR_EXTERNAL(CreateCompatibleDC)(hDCWindow);
	if (hDCAlphaBuffer == NULL)
	{
		p_timeEndPeriod(1);
		ExitProcess(1);
	}
	hBitmapAlphaBuffer = PTR_EXTERNAL(CreateCompatibleBitmap)(hDCWindow, size.x, size.y);
	if (hBitmapAlphaBuffer == NULL)
	{
		p_timeEndPeriod(1);
		ExitProcess(1);
	}
	hOldBitmapAlphaBuffer = (HBITMAP)PTR_EXTERNAL(SelectObject)(hDCAlphaBuffer, hBitmapAlphaBuffer);
	PTR_EXTERNAL(SetBkMode)(hDCAlphaBuffer, TRANSPARENT);
#endif

	hDCBackground = PTR_EXTERNAL(CreateCompatibleDC)(hDCWindow);
	if (hDCBackground == NULL)
	{
		p_timeEndPeriod(1);
		ExitProcess(1);
	}
	hBitmapBackground = PTR_EXTERNAL(CreateCompatibleBitmap)(hDCWindow, size.x, size.y);
	if (hBitmapBackground == NULL)
	{
		p_timeEndPeriod(1);
		ExitProcess(1);
	}
	hOldBitmapBackground = (HBITMAP)PTR_EXTERNAL(SelectObject)(hDCBackground, hBitmapBackground);
	FixStretching(hDCBackground);

	if (hDCBackgroundBB == NULL)
	{
		hDCBackgroundBB = PTR_EXTERNAL(CreateCompatibleDC)(hDCWindow);
		if (hDCBackgroundBB == NULL)
		{
			p_timeEndPeriod(1);
			ExitProcess(1);
		}
		hBitmapBackgroundBB = PTR_EXTERNAL(CreateCompatibleBitmap)(hDCWindow, Gradient::PIC_SIZE_X, Gradient::PIC_SIZE_Y);
		if (hBitmapBackgroundBB == NULL)
		{
			p_timeEndPeriod(1);
			ExitProcess(1);
		}
		hOldBitmapBackgroundBB = (HBITMAP)PTR_EXTERNAL(SelectObject)(hDCBackgroundBB, hBitmapBackgroundBB);
		FixStretching(hDCBackgroundBB);

		BlitBackgroundBB();
	}
	BlitBackground();
}

bool IsCreated()
{
	bool result;
	{
		Synchronized sync1(lockGUI);
		result = created;
	}
	return result;
}

//
//  FUNCTION: WndProc(HWND, unsigned, WORD, LONG)
//
//  PURPOSE:  Processes messages for the main window.
//
//
LRESULT CALLBACK WndProc(
	HWND inHWnd,
	UINT message,
	WPARAM wParam,
	LPARAM lParam
)
{
	switch (message)
	{
	case WM_KEYDOWN:
		{
			Synchronized sync1(lockGUI);

			int key = int(wParam);
			if (!field.IsGameOver())
			{
				if (key == VK_P)
				{
					if (paused)
					{
						timerIndex = savedTimerIndex;
					}
					else
					{
						savedTimerIndex = timerIndex;
					}
					paused = !paused;
				}
				else if (paused)
				{
					// do nothing.
				}
				else if (
					key == VK_LEFT ||
					key == VK_NUMPAD4 ||
					key == VK_J ||
					key == VK_A
				)
				{
					if (field.MoveLeft())
					{
						movementStageLeft.Escalate();
					}
				}
				else if (
					key == VK_RIGHT ||
					key == VK_NUMPAD6 ||
					key == VK_L ||
					key == VK_D
				)
				{
					if (field.MoveRight())
					{
						movementStageRight.Escalate();
					}
				}
				else if (
					key == VK_UP ||
					key == VK_NUMPAD8 ||
					key == VK_I ||
					key == VK_W
				)
				{
					field.Rotate();
				}
				else if (
					key == VK_OTHER_KP5 ||
					key == VK_NUMPAD5 ||
					key == VK_K ||
					key == VK_S
				)
				{
					if (field.MoveDown())
					{
						movementStageDownwards.Escalate();
					}
				}
				else if (
					key == VK_SPACE ||
					key == VK_DOWN ||
					key == VK_NUMPAD2
				)
				{
					if (field.Drop())
					{
						timerIndex = 0;
					}
				}
			}
			else
			{
				if (key == VK_RETURN)
				{
					paused = false;
					field.NewGame();
					timerIndex = 0;
					movementStageLeft.Cancel();
					movementStageRight.Cancel();
					movementStageDownwards.Cancel();
				}
			}
		}
		break;
	case WM_CREATE:
		{
			Synchronized sync0(lockBackground);
			Synchronized sync1(lockGUI);
			black = PTR_EXTERNAL(CreateSolidBrush)(RGB(0, 0, 0));
			white = PTR_EXTERNAL(CreateSolidBrush)(RGB(255, 255, 255));
			textFont = PTR_EXTERNAL(CreateFont)(15, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE,
				DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
				DEFAULT_QUALITY, FIXED_PITCH | FF_DONTCARE, TEXT("Courier New"));
			DeleteMemDC();
			DeleteBackgroundBB();
			PostMessage(inHWnd, WMU_FINISH, 0, 0);
		}
		break;
	case WM_DESTROY:
		{
			// All cleanup code removed
			p_timeEndPeriod(1);
			ExitProcess(0);
		}
		break;
	case WM_DEVMODECHANGE:
		// fall through:
	case WM_SIZE:
		{
			Synchronized sync0(lockBackground);
			Synchronized sync1(lockGUI);
			if (created)
			{
				DeleteMemDC();
				CreateMemDC();
			}
		}
		break;
	case WM_ERASEBKGND:
		return 1;
	case WM_PAINT:
		{
			PAINTSTRUCT paintStruct;
			BeginPaint(inHWnd, &paintStruct);
			EndPaint(inHWnd, &paintStruct);
		}
		break;
	case WMU_FINISH:
		{
			Synchronized sync0(lockBackground);
			Synchronized sync1(lockGUI);
			CreateMemDC();
			created = true;
		}
		break;
	default:
		return DefWindowProc(inHWnd, message, wParam, lParam);
	}
	return 0;
}

//
//  FUNCTION: GradientPositionsThread(LPVOID)
//
//  PURPOSE:  Time the gradient background thread.
//            This thread changes the locations of the screen spots.
//
//
static DWORD WINAPI GradientPositionsThread(LPVOID /*lpParameter*/)
{
	static const DWORD SLEEP_TIME = 1 * 1000;
	static double m_step_x[GradientChange::INITIAL_POINTS];
	static double m_step_y[GradientChange::INITIAL_POINTS];
	static int m_steps[GradientChange::INITIAL_POINTS];

	int i;

	ZeroMemory(m_steps, sizeof (m_steps));

	Timer timer(SLEEP_TIME);

	timer.Start();

	for (;;)
	{
		for (i = 0; i < GradientChange::INITIAL_POINTS; i++)
		{
			if (m_steps[i] == 0)
			{
				m_steps[i] = MathRound<int>(MathRandom() * 200) + 51 +
#ifdef GRADIENT_STARS_UP
					(INITIAL_HEIGHT / 2)
#else
					INITIAL_HEIGHT
#endif
					;
				double targetX = MathRandom();
#ifdef GRADIENT_STARS_UP
				double targetY = 1.0 - (MathRandom() / 2.0);
#else
				double targetY = MathRandom();
#endif
				double dSteps = double(m_steps[i]);
				m_step_x[i] = (targetX - gradientChange.m_points_x[i]) / dSteps;
				m_step_y[i] = (targetY - gradientChange.m_points_y[i]) / dSteps;
			}
			--m_steps[i];
			gradientChange.m_points_x[i] += m_step_x[i];
			gradientChange.m_points_y[i] += m_step_y[i];
		}
		gradientChange.Init();
		{
			Synchronized sync0(lockBackground);
			CopyMemory(&gradient.m_change, &gradientChange, sizeof (GradientChange));
		}
		timer.Proceed();
	}
	// Unreachable code:
    #ifdef __GNUC__
	return 0;
    #endif
}

//
//  FUNCTION: GradientColorsThread(LPVOID)
//
//  PURPOSE:  Time the gradient background thread.
//            This thread changes the colors of the screen spots.
//
//
static DWORD WINAPI GradientColorsThread(LPVOID /*lpParameter*/)
{
	static const __int64 SLEEP_TIME = 1 * 1000;

	Timer timer(SLEEP_TIME);

	timer.Start();

	for (;;)
	{
		{
			Synchronized sync0(lockBackground);
			gradient.ReInit();
			gradient.Fill();
		}
		timer.Proceed();
		{
			Synchronized sync0(lockBackground);
			BlitBackgroundBB();
			{
				Synchronized sync2(lockMutual);
				BlitBackground();
			}
		}
	}
	// Unreachable code:
    #ifdef __GNUC__
	return 0;
    #endif
}

static void PaintSquares(bool paintFalling, bool transitions)
{
#ifdef SMOOTH_FALLING
	const float timerShare = float(paused ? savedTimerIndex : timerIndex) / float(TIMER_INDEX);
#endif

	float colWidth = (float)size.x / (float)Field::GRID_COLS;
	float rowHeight = (float)size.y / (float)Field::GRID_ROWS;
	float y = 0;
	int iy = 0;

	int row, col;

	float nexty;
	int nextiy;
	int h;
	float x;
	int ix;

	float nextx;
	int nextix;
	Square * s;
	bool falling;

	int hDiff = 0;

#ifdef SMOOTH_FALLING
	int smoothFallingDiff = MathRound<int>(rowHeight - timerShare * rowHeight);
#endif

	for (row = 0; row < Field::GRID_ROWS; ++row)
	{
		nexty = y + rowHeight;
		nextiy = MathRound<int>(nexty);
		h = nextiy - iy;
		x = 0;
		ix = 0;
		for (col = 0; col < Field::GRID_COLS; ++col)
		{
			nextx = x + colWidth;
			nextix = MathRound<int>(nextx);
			s = field.GetSquare(row, col);
			falling = field.IsFalling(row, col);

			if (falling == paintFalling && s != NULL)
			{
				if (falling && transitions)
				{
					hDiff = 0;
#ifdef SMOOTH_FALLING
					hDiff += smoothFallingDiff;
#endif
#ifdef SMOOTH_ACCELERATE
					hDiff += movementStageDownwards.GetValue(h);
#endif
					s->PaintRectangle(
						ix
#ifdef SMOOTH_ACCELERATE
							+ movementStageLeft.GetValue(w)
							- movementStageRight.GetValue(w)
#endif
						,
						iy - hDiff,
						nextix - ix,
						h);
				}
				else if (falling)
				{
					hDiff = 0;
#ifdef SMOOTH_FALLING
					hDiff += smoothFallingDiff;
#endif
					s->PaintRectangle(ix, iy - hDiff, nextix - ix, h);
				}
				else
				{
					s->PaintRectangle(ix, iy, nextix - ix, h);
				}
			}
			x = nextx;
			ix = nextix;
		}
		y = nexty;
		iy = nextiy;
	}

}

static void TickTetris()
{
	starField.Move();
	starField.Create(&size);

	timerIndex++;
	if (timerIndex == TIMER_INDEX)
	{
		if (!field.IsGameOver())
		{
			if (!paused)
			{
				if (!field.MoveDownward())
				{
					movementStageLeft.Cancel();
					movementStageRight.Cancel();
					movementStageDownwards.Cancel();
				}
#ifdef BEEP_MOVE
				Beep(/*750*/ 300 , 10);
#endif
			}
		}
		timerIndex = 0;
	}
	HBRUSH oldBrush = (HBRUSH)PTR_EXTERNAL(SelectObject)(hDCGlobalBackBuffer, black);
	/////////////////////////////
	// Background:
	/////////////////////////////

	SetClip(0, 0, size.x, size.y);
	{
		Synchronized sync2(lockMutual);
		PTR_EXTERNAL(BitBlt)(hDCGlobalBackBuffer, 0, 0, size.x, size.y, hDCBackground, 0, 0, SRCCOPY);
	}
	starField.Paint();

	/////////////////////////////

#ifdef SMOOTH_ACCELERATE
	{
		RECT rect;
		rect.right = size.x;
		rect.bottom = size.y;
		rect.left = 0;
		rect.top = 0;
		PTR_EXTERNAL(BitBlt)(hDCAlphaBuffer, 0, 0, size.x, size.y, hDCGlobalBackBuffer, 0, 0, SRCCOPY);
	}
	HDC dc = hDCGlobalBackBuffer;
	hDCGlobalBackBuffer = hDCAlphaBuffer;
	PaintSquares(true, false);
	hDCGlobalBackBuffer = dc;
	AlphaBlend(
		hDCGlobalBackBuffer, 0, 0, size.x, size.y, hDCAlphaBuffer, 0, 0, size.x, size.y, blendFunction);
#endif
	PaintSquares(false, false);
	PaintSquares(true, true);

	/////////////////////////////
	int lineCount = field.GetLineCount();
	if (lineCount != oldLineCount)
	{
		oldLineCount = lineCount;
		ZeroMemory(lineCountBuf, sizeof(lineCountBuf));
		_sntprintf(lineCountBuf,
			ELEMENTS(lineCountBuf) - 1,
			TEXT("Lines : %d"), lineCount);
	}

	DrawString(lineCountBuf, HORIZONTAL_TEXT_START, VERTICAL_TEXT_START);

	/////////////////////////////
	if (field.IsGameOver())
	{
        const TCHAR * const gameOver = TEXT("Game Over");
		DrawString(gameOver, HORIZONTAL_TEXT_START, VERTICAL_TEXT_START + 12);
        const TCHAR * const playAgain = TEXT("Play Again : [ENTER]");
		DrawString(playAgain, HORIZONTAL_TEXT_START, VERTICAL_TEXT_START + 12 + 12);
	}
	else if (paused)
	{
		DrawString(TEXT("Paused"), HORIZONTAL_TEXT_START, VERTICAL_TEXT_START + 12);
		DrawString(TEXT("Continue : [P]"), HORIZONTAL_TEXT_START, VERTICAL_TEXT_START + 12 + 12);
	}

	PTR_EXTERNAL(SelectObject)(hDCGlobalBackBuffer, oldBrush);
	PTR_EXTERNAL(BitBlt)(hDCWindow, 0, 0, size.x, size.y, hDCGlobalBackBuffer, 0, 0, SRCCOPY);

	if (!paused)
	{
		movementStageLeft.Advance();
		movementStageRight.Advance();
		movementStageDownwards.Advance();
	}
}

//
//  FUNCTION: TetrisTickProc(LPVOID)
//
//  PURPOSE:  Time the Tetris game.
//            This thread ticks the game and paints the window.
//            It also starts the two gradient threads.
//
//
static DWORD WINAPI TetrisTickProc(LPVOID /*lpParameter*/)
{
	static const __int64 SLEEP_TIME = 30;

	Timer timer(SLEEP_TIME);

	while (!IsCreated())
	{
		Sleep(1);
	}
	{
		DWORD threadID;
		HANDLE threadHandle;

		Synchronized sync0(lockBackground);
		Synchronized sync1(lockGUI);
		threadHandle = CreateThread(
			NULL,
			0,
			(LPTHREAD_START_ROUTINE)GradientColorsThread,
			0,
			0,
			&threadID
			);
		SetThreadPriority(threadHandle, THREAD_PRIORITY_BELOW_NORMAL);
		threadHandle = CreateThread(
			NULL,
			0,
			(LPTHREAD_START_ROUTINE)GradientPositionsThread,
			0,
			0,
			&threadID
			);
		SetThreadPriority(threadHandle, THREAD_PRIORITY_LOWEST);
	}
	timer.Start();
	for (;;)
	{
		timer.Proceed();
		{
			Synchronized sync1(lockGUI);
			TickTetris();
		}
	}
	// Unreachable code:
    #ifdef __GNUC__
	return 0;
    #endif
}
